/********************************************************                                          
                                           2009 1/1 
                                          2011 3/19 update
  Linux KeyHoleTV (C) 2008 2009           OISEYER Inc.
                                        Takashi Kosaka
   Linux DSP Sound Dynamic Link Library

   File:                 DSPObj.c
   Dynamic Link Library: DSPObj.so

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   
   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <unistd.h>
#include <sys/select.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <time.h>
#include <unistd.h>
#include <linux/soundcard.h>

#define   MAX_AUDIO_NUM                 20
#define   AUDIO_44100_100_NUM         10240    /* 1000 mill sec 44100 Hz Bytes */
#define   AUDIO_NOTIFY_SIZE_2         4410     /* 16 bits 44100 Hz 100 Mill Sec short int */


/*****************************************************************
  Sound Object
  Interface structure (Do not change)
*****************************************************************/
typedef struct _SoundObj {
	void        *Sample;  // used by Linux KeyHoleTV 
	void        *Handle;

	unsigned char *(*GetAudioBuffer)(void *OO);
	void           (*SetNextWriteBuffer)(void *OO,int Size);
	void           (*SoundFunction)(void *OO);
	void           (*StopAuidoFunction)(void *OO);
	void           (*FinishAuidoFunction)(void *OO);

} DataSoundObj, *PSoundObj;

/**********************************************************
 Audio Buffer
***********************************************************/
typedef struct _AudioBuffer {
  unsigned char *Buf;   
  int            Size;
  int            Written;
  struct _AudioBuffer *Next;
} AudioBuffer, *PAudioBuffer;

/************************************************************
  DSP Object
************************************************************/
typedef struct _DSPObj {
  int          Fd;
  int          StopIDW;
  int          StopIDR;

  unsigned int Store;
  unsigned int Used;
  int          Start;
  int          DspSize;
	
  PAudioBuffer Read;
  PAudioBuffer Write;
	char         *Exiting;
	char         *JoinRoom;
} DSPObj, *PDSPObj;

/*************************************************************
   Set DSP Driver
*************************************************************/
static int SetDSP(PDSPObj Obj)
{
  int fmt     = AFMT_S16_LE;
  int freq    = 44100;
  int channel = 1;
  audio_buf_info    Info;

  Obj->Fd = -1;

  if( (Obj->Fd = open("/dev/dsp",O_WRONLY)) < 0 ) {
    return(-1);
  }

  if(ioctl(Obj->Fd, SOUND_PCM_RESET, 0 ) < 0 ) {
    return(-1);
  }

  /* $B%5%&%s%I%U%)!<%^%C%H$N@_Dj(B */
  if ( ioctl( Obj->Fd, SOUND_PCM_SETFMT, &fmt ) == -1 ) {
    return(-1);
  }
		
  /* $B%A%c%M%k?t$N@_Dj(B */
  if ( ioctl( Obj->Fd, SOUND_PCM_WRITE_CHANNELS, &channel ) == -1 )  {
    return(-1);
  }

  /* $B%5%s%W%j%s%0<~GH?t$N@_Dj(B */
  if ( ioctl(Obj->Fd, SOUND_PCM_WRITE_RATE, &freq ) == -1 ){
    return(-1);
  }
		
  if( ioctl(Obj->Fd, SNDCTL_DSP_GETOSPACE, &Info) == 0 ) {
    Obj->DspSize = Info.bytes;
  }
  else {
    Obj->DspSize = 100000;
  }

  return(0);
}

/********************************************************
  Store Last Sound Data
********************************************************/
static int LastSoundData = 0;

/**********************************************************
  Write NULL Sound
**********************************************************/
static void WriteNullSound(short *Buf,int aaa)
{
  int i;
  short int DD;
  double Sa,LastD;

  aaa = aaa >> 2;
  
  if( LastSoundData == 0 ) {
    for( i = 0 ; i < aaa ; i++ ) {
      *(Buf + i) = 0;
    }
    return;
  }

  LastD = (double)LastSoundData;
  Sa = (double)abs(LastSoundData) / (double)aaa;

  //fprintf(stderr,"Last %d Sa:%e\n",LastSoundData,Sa);

  if( LastSoundData > 0 ) {
    LastD = -LastD + Sa;
  }
  else {
    LastD = -LastD - Sa;
  }
  
  for( i = 0 ; i < aaa ; i++ ) {
    DD = (short)LastD;
    *(Buf + i) = DD;

    if( LastD > 0.0 ) {
      LastD = -LastD + Sa;
    }
    else if (LastD < 0.0 ) {
      LastD = -LastD - Sa;
    }
  }
  LastSoundData = 0;
}


/************************************************************
   Write Sound to ESD Driver
************************************************************/
static void WriteSoundToDspDriver(PDSPObj Obj)
{
  fd_set            WriteFd;
  int               Ret,W,bb,Size;
  struct            timeval TT;
  audio_buf_info    Info;

  if( Obj->Fd > 0 ) {
    short Buf[AUDIO_44100_100_NUM];

    FD_ZERO(&WriteFd);
    TT.tv_sec  = 0;
    TT.tv_usec = 0;
    W = Obj->Fd + 1;
    FD_SET(Obj->Fd,&WriteFd);

    if( Obj->Read != Obj->Write ) {

      if((Ret = select(W,NULL,&WriteFd,NULL,&TT)) > 0 ) {

	if( FD_ISSET(Obj->Fd,&WriteFd) ) { /* $B=q$-9~$_5v2D(B */

	  if( ioctl(Obj->Fd, SNDCTL_DSP_GETOSPACE, &Info) == 0 ) {
	    if( Info.bytes >= AUDIO_NOTIFY_SIZE_2 ) {

	      if( Info.bytes > (Obj->Read->Size - Obj->Read->Written )) {
		Size = Obj->Read->Size - Obj->Read->Written;
	      }
	      else {
		Size = Info.bytes;
	      }
							
	      bb = write(Obj->Fd,Obj->Read->Buf + Obj->Read->Written,Size);

	      if( bb > 0 ) {
		Obj->Read->Written += bb;
		if( Obj->Read->Written >= Obj->Read->Size ) {
		  Obj->Used++;
		  Obj->Read->Written = Obj->Read->Size = 0;
		  Obj->Read = Obj->Read->Next;
		}
	      }
	    }
	  }
	}
      }
    }
    else {
      if((Ret = select(W,NULL,&WriteFd,NULL,&TT)) > 0 ) {
	
	if( FD_ISSET(Obj->Fd,&WriteFd) ) { /* $B=q$-9~$_5v2D(B */
	  
	  if( ioctl(Obj->Fd, SNDCTL_DSP_GETOSPACE, &Info) == 0 ) {
	    if( Info.bytes >= AUDIO_NOTIFY_SIZE_2 ) {
	      Size = Info.bytes;
	      /* Create Null Data */
	      WriteNullSound(Buf,Size);
	      write(Obj->Fd,(char *)Buf,Size);
	    }
	  }
	}
      }
#if 0
      if( ioctl(Obj->Fd, SNDCTL_DSP_GETOSPACE, &Info) == 0 ) {
	if( Info.bytes >= (Obj->DspSize - 500) ) {
	  /*$B%P%C%U%!$,$+$i(B */
	  close(Obj->Fd);
	  Obj->Fd = -1;
	  Obj->Start = 0;
	}
      }
#endif
    }
  }
}

		
/************************************************************
  Interface Functions
************************************************************/

/************************************************************
  Initialize DSP Driver And Allocate Memory
  Call by GTK Main Thread
************************************************************/
static void * InitDspDriver(char *Exiting,char *JoinRoom)
{
  PDSPObj      Obj;
  int          i;
  PAudioBuffer Now,Prv,Top;		
	
  if((Obj = (PDSPObj)malloc(sizeof(DSPObj))) == NULL ) {
    return((void *)NULL);
  }

  if( SetDSP(Obj) < 0 ) {
    return((void *)NULL);
  }
	Obj->Exiting  = Exiting;
	Obj->JoinRoom = JoinRoom;

  Obj->StopIDW = Obj->StopIDR = 0;
  Obj->Store = Obj->Used = Obj->Start = 0;

  Top = Prv = NULL;

  for( i = 0 ; i < MAX_AUDIO_NUM ; i++ ) {
    if((Now = (PAudioBuffer)malloc(sizeof(AudioBuffer))) == NULL ) {
      return((void *)NULL);
    }

    if((Now->Buf = (unsigned char *)malloc(AUDIO_44100_100_NUM)) == NULL ) {
      return((void *)NULL);
    }

    Now->Size = Now->Written = 0;

    if( ! Top ) {
      Top = Now;
    }

    if( Prv ) {
      Prv->Next = Now;
    }

    Prv = Now;
  }
  Now->Next = Top;

  /* Store Ring Buffer */
  Obj->Read = Obj->Write = Now;

  return((void *)Obj);
}

/***********************************************************
   Get Wirte Buffer
   Called by Network Thread
***********************************************************/
static unsigned char *GetWriteDSPBuffer(void *OO)
{
  PDSPObj            Obj;
  Obj = (PDSPObj)OO;
  return((unsigned char *)Obj->Write->Buf);
}

/**********************************************************
  Set Next Write Buffer
  Called by Network Thread
**********************************************************/
static void SetNextDSPWriteBuffer(void *OO,int Size)
{
  PDSPObj            Obj;
  Obj = (PDSPObj)OO;
  Obj->Write->Size = Size;
  Obj->Write->Written = 0;
  Obj->Write = Obj->Write->Next;
  Obj->Store++;
}

/**********************************************************
  Sound Function
  Called by sound Thread by g_thread_create (Each 10 mill sec 
  the thread calls this function)
**********************************************************/
static void SoundDSPFunction(void *OO)
{
  PDSPObj            Obj;
  Obj = (PDSPObj)OO;

  if( Obj->StopIDW != Obj->StopIDR ) {

    /* $BDd;_MW5a$,Mh$?(B */
    if( Obj->Fd > 0 ) {
      close(Obj->Fd);
      Obj->Fd = -1;
    }
		
    Obj->Read = Obj->Write;
    Obj->Read->Size = Obj->Read->Written = 0;
    Obj->Start   = 0;
    Obj->StopIDR = Obj->StopIDW;
    Obj->Used    = Obj->Store = 0;
  }

  if( *Obj->Exiting == 0 && *Obj->JoinRoom ) {
    if( Obj->Fd < 0 ) {
      SetDSP(Obj);
    }

    if(  Obj->Start == 0 ) {	
      if( (Obj->Used + 5) < Obj->Store  ) {	
	Obj->Start = 1;
	WriteSoundToDspDriver(Obj);
      }
    }
    else {
      WriteSoundToDspDriver(Obj);
    }
  }
}

/********************************************************
  Stop DSP Audio
  Called by GTK Main Thread
********************************************************/
static void StopDSPAudio(void *OO)
{
  PDSPObj            Obj;
  Obj = (PDSPObj)OO;
  Obj->StopIDW++;
}

/********************************************************
  Finish DSP Audio
  Called by GTK Main Thread
********************************************************/
static void FinishDSPAudio(void *OO)
{
  PDSPObj            Obj;
  PAudioBuffer        Now,Next;
  int                 i;
  Obj = (PDSPObj)OO;

  if( Obj->Fd > 0 ) {
    close(Obj->Fd);
  }
  Now = Obj->Read;
  for( i = 0 ; i < MAX_AUDIO_NUM ; i++ ) {
    Next = Now->Next;
    free((void *)Now);
    Now = Next;
  }
  free((void *)Obj);
}

/***************************************************************
   Set DSP Driver
  This function called by dlsym(Handle,"SetAudioDriver");
  So, propgramers can not change function mame and arguments.
  Called by GTK Main Thread.
***************************************************************/
int SetAudioDriver(PSoundObj Obj,char *exiting,char *Join)
{
  if((Obj->Handle = InitDspDriver(exiting,Join)) == NULL ) {
    return(1);
  }

	Obj->GetAudioBuffer     = GetWriteDSPBuffer;
  Obj->SetNextWriteBuffer = SetNextDSPWriteBuffer;
  Obj->SoundFunction      = SoundDSPFunction;
  Obj->StopAuidoFunction  = StopDSPAudio;
  Obj->FinishAuidoFunction = FinishDSPAudio;

  return(0);
}
